*+====================================================================
*+
*+    Source Module => D:\SRC\USCOPE\USCOPE.PRG
*+
*+    USCOPE.PRG, Public Domain by Phil Barnett
*+    
*+    USCOPE2.PRG, Public Domain by Roger J. Donnay
*+    
*+    Universal Scope Functions for Clipper 5.x, Xbase++
*+
*+    Functions: Procedure U_SETSCOPE()
*+               Procedure U_CLRSCOPE()
*+               Procedure U_DBGOTOP()
*+               Procedure U_DBGOBOTTOM()
*+               Function U_BOF()
*+               Function U_EOF()
*+               Function U_DBSeekLast()
*+               Static Procedure U_ScopeFocus()
*+
*+    Reformatted by Click! 1.13b4 on May-31-1998 at  1:56 pm
*+
*+====================================================================

/*

Version 1.1 Added static array destruct for unused scopes & aliases

            Speedup of U_EOF() and U_BOF() by Sean Webb

            Calling U_CLRSCOPE() with no parameters clears both scopes

            Reduced dgroup usage to 1 static array per Sean Webb

            Implemented U_ScopeFocus bypass suggested by Sean Webb

            More corrections to U_EOF and U_BOF per Sean Webb

*/

static aScopeData := { { 0, 0, '@@@', 0, '+++', 0 } }

#define nAliasPointer aScopeData[ 1, 1 ]
#define nIndexPointer aScopeData[ 1, 2 ]
#define cAliasName    aScopeData[ 1, 3 ]
#define nIndexNumber  aScopeData[ 1, 4 ]
#define cLastAlias    aScopeData[ 1, 5 ]
#define nLastIndex    aScopeData[ 1, 6 ]
#define lTopScopeSet  aScopeData[ nAliasPointer, 2, nIndexPointer, 2, 1 ]
#define lBotScopeSet  aScopeData[ nAliasPointer, 2, nIndexPointer, 2, 2 ]
#define uTopScope     aScopeData[ nAliasPointer, 2, nIndexPointer, 2, 3 ]
#define uBotScope     aScopeData[ nAliasPointer, 2, nIndexPointer, 2, 4 ]

#define VERSION 1.1

*+--------------------------------------------------------------------
*+
*+    Procedure U_SETSCOPE()
*+
*+--------------------------------------------------------------------
*+
procedure U_SETSCOPE( nMode, uThisScope )

U_ScopeFocus()

do case
case nMode == 0
   lTopScopeSet := .t.
   uTopScope    := uThisScope
case nMode == 1
   lBotScopeSet := .t.
   uBotScope    := uThisScope
endcase

return

*+--------------------------------------------------------------------
*+
*+    Procedure U_CLRSCOPE()
*+
*+--------------------------------------------------------------------
*+
procedure U_CLRSCOPE( nMode )

local x
local y

U_ScopeFocus()

do case
case nMode == NIL
   lTopScopeSet := .f.
   lBotScopeSet := .f.
case nMode == 0
   lTopScopeSet := .f.
   uTopScope    := NIL
case nMode == 1
   lBotScopeSet := .f.
   uBotScope    := NIL
endcase

if !lTopScopeSet .and. !lBotScopeSet
   // invalidate bypass
   cLastAlias   := '@@@'
   cAliasName   := '+++'
   nLastIndex   := 0
   nIndexNumber := 0
endif

// destroy any array elements containing empty scopes
if !lTopScopeSet .and. !lBotScopeSet .and. len( aScopeData ) > 1

   for x := len( aScopeData ) to 2
      for y := len( aScopeData[ x, 2 ] ) to 1 step - 1
         if !aScopeData[ x, 2, y, 2, 1 ] .and. !aScopeData[ x, 2, y, 2, 2 ]
            adel( aScopeData[ x, 2 ], x )
            asize( aScopedata[ x, 2 ], len( aScopedata[ x, 2 ] ) - 1 )
         endif
      next
      if len( aScopeData[ x, 2 ] ) == 0
         adel( aScopeData, x )
         asize( aScopeData, len( aScopeData ) - 1 )
      endif
   next

endif

return

*+--------------------------------------------------------------------
*+
*+    Procedure U_DBGOTOP()
*+
*+--------------------------------------------------------------------
*+
procedure U_DBGOTOP()

U_ScopeFocus()

if lTopScopeSet
   dbseek( uTopScope, .t. )             // softseek is on
else
   dbgotop()
endif

return

*+--------------------------------------------------------------------
*+
*+    Procedure U_DBGOBOTTOM()
*+
*+--------------------------------------------------------------------
*+
procedure U_DBGOBOTTOM()

U_ScopeFocus()

if lBotScopeSet
   U_DBSeekLast( uBotScope )
else
   dbgobottom()
endif

return

*+--------------------------------------------------------------------
*+
*+    Function U_BOF()
*+
*+--------------------------------------------------------------------
*+
function U_BOF()

local lBof      := bof()
local cIndexKey

U_ScopeFocus()

if !lBof .and. lTopScopeSet .and. !empty( cIndexKey := indexkey( 0 ) )
   lBof := uTopScope > &( cIndexKey )
endif

return lBof

*+--------------------------------------------------------------------
*+
*+    Function U_EOF()
*+
*+--------------------------------------------------------------------
*+
function U_EOF()

local lEof      := eof()
local cIndexKey

U_ScopeFocus()

if !lEof .and. lBotScopeSet .and. !empty( cIndexKey := indexkey( 0 ) )
   lEof := uBotScope < &( cIndexKey )
endif

return lEof

*+--------------------------------------------------------------------
*+
*+    Function U_DBSeekLast()
*+
*+    Called from ( uscope.prg   )   1 - procedure u_dbgobottom()
*+
*+--------------------------------------------------------------------
*+
function U_DBSeekLast( uVariable )

// The basis for this seeklast() example was included in the Xbase++
// documentation under dbseek() sample code.

// It is important to know that the original version of dbseeklast only
// worked for character strings.

local cIndexKey
local nLen
local lRetVal   := .f.

U_ScopeFocus()

begin sequence

   cIndexKey := indexkey( 0 )
   if empty( cIndexKey )                // no index active
      break
   endif

   do case
   case valtype( uVariable ) == 'C'

      // increase last Chr() by 1 for SOFTSEEK
      nLen := len( uVariable )
      // SOFTSEEK ON !
      dbseek( left( uVariable, nLen - 1 ) + chr( asc( right( uVariable, 1 ) ) + 1 ), .T. )

      // skip back 1
      dbskip( - 1 )

      // determine value of index key
      if ( uVariable == left( &( cIndexKey ), nLen ) )
         // match found!
         lRetVal := .t.
         break
      endif

   case valtype( uVariable ) == 'D'

      // increase date by 1 for SOFTSEEK
      dbseek( uVariable + 1, .T. )

      // skip back 1
      dbskip( - 1 )

      // determine value of index key
      if uVariable == &( cIndexKey )
         // match found!
         lRetVal := .t.
         break
      endif

   case valtype( uVariable ) == 'N'

      // increase numeric by 1 for SOFTSEEK
      dbseek( uVariable + 1, .T. )

      // skip back 1
      dbskip( - 1 )

      // determine value of index key
      if uVariable == &( cIndexKey )
         // match found!
         lRetVal := .t.
         break
      endif

   otherwise
      break
   endcase

   // if Eof() was .T.
   dbskip( 1 )

end sequence

return lRetVal

*+--------------------------------------------------------------------
*+
*+    Static Procedure U_ScopeFocus()
*+
*+    Called from ( uscope.prg   )   1 - procedure u_setscope()
*+                                   1 - procedure u_clrscope()
*+                                   1 - procedure u_dbgotop()
*+                                   1 - procedure u_dbgobottom()
*+                                   1 - function u_bof()
*+                                   1 - function u_eof()
*+                                   1 - function u_dbseeklast()
*+
*+--------------------------------------------------------------------
*+
static procedure U_ScopeFocus()

local nPointer
local cAlias
local nIndex   := indexord()

if indexord() = 0
   // invalidate bypass
   cLastAlias   := '@@@'
   cAliasName   := '+++'
   nLastIndex   := 0
   nIndexNumber := 0
endif

   cAlias := alias()

   // check bypass condition
   if cAlias <> cLastAlias .or. nIndex <> nLastIndex

      // remember to start ascan on element 2. Element 1 is the bypass info

      nPointer := ascan( aScopeData, { | x | x[ 1 ] == cAlias }, 2 )

      if nPointer == 0
         aadd( aScopeData, { cAlias, {} } )
         nAliasPointer := len( aScopeData )
      else
         nAliasPointer := nPointer
      endif

      if len( aScopeData[ nAliasPointer, 2 ] ) > 0
         nPointer := ascan( aScopeData[ nAliasPointer, 2 ], { | x | x[ 1 ] == nIndex } )
      else
         nPointer := 0
      endif

      if nPointer == 0
         aadd( aScopeData[ nAliasPointer, 2 ], { nIndex, { .f., .f., NIL, NIL } } )
         nIndexPointer := len( aScopeData[ nAliasPointer, 2 ] )
      else
         nIndexPointer := nPointer
      endif

      // synchronize bypass variables
      cLastAlias := cAlias
      nLastIndex := nIndex

   endif

return

*+ EOF: USCOPE.PRG
